@@ -1,9 +1,9 @@
 /*****************************************************************************
 
-Copyright (c) 1996, 2017, Oracle and/or its affiliates. All rights reserved.
+Copyright (c) 1996, 2016, Oracle and/or its affiliates. All rights reserved.
 Copyright (c) 2008, Google Inc.
 Copyright (c) 2009, Percona Inc.
-Copyright (c) 2017, MariaDB Corporation.
+Copyright (c) 2013, 2015, MariaDB Corporation
 
 Portions of this file contain modifications contributed and copyrighted by
 Google, Inc. Those modifications are gratefully acknowledged and are described
@@ -52,6 +52,7 @@ Created 2/16/1996 Heikki Tuuri
 #include "os0file.h"
 #include "os0thread.h"
 #include "fil0fil.h"
+#include "fil0crypt.h"
 #include "fsp0fsp.h"
 #include "rem0rec.h"
 #include "mtr0mtr.h"
@@ -68,11 +69,19 @@ Created 2/16/1996 Heikki Tuuri
 #include "ibuf0ibuf.h"
 #include "srv0start.h"
 #include "srv0srv.h"
+#include "buf0flu.h"
+#include "btr0defragment.h"
+#include "ut0timer.h"
+#include "btr0scrub.h"
+
+#include <mysql/service_wsrep.h>
+
 #ifndef UNIV_HOTBACKUP
 # include "trx0rseg.h"
 # include "os0proc.h"
 # include "sync0sync.h"
 # include "buf0flu.h"
+# include "buf0mtflu.h"
 # include "buf0rea.h"
 # include "dict0boot.h"
 # include "dict0load.h"
@@ -112,9 +121,6 @@ UNIV_INTERN ibool	srv_have_fullfsync = FALSE;
 /** TRUE if a raw partition is in use */
 UNIV_INTERN ibool	srv_start_raw_disk_in_use = FALSE;
 
-/** UNDO tablespaces starts with space id. */
-ulint	srv_undo_space_id_start;
-
 /** TRUE if the server is being started, before rolling back any
 incomplete transactions */
 UNIV_INTERN ibool	srv_startup_is_before_trx_rollback_phase = FALSE;
@@ -130,7 +136,7 @@ SRV_SHUTDOWN_CLEANUP and then to SRV_SHUTDOWN_LAST_PHASE, and so on */
 UNIV_INTERN enum srv_shutdown_state	srv_shutdown_state = SRV_SHUTDOWN_NONE;
 
 /** Files comprising the system tablespace */
-static pfs_os_file_t	files[1000];
+static os_file_t	files[1000];
 
 /** io_handler_thread parameters for thread identification */
 static ulint		n[SRV_MAX_N_IO_THREADS];
@@ -146,7 +152,10 @@ SRV_MAX_N_IO_THREADS + 7: srv_worker_thread
 ...
 SRV_MAX_N_IO_THREADS + 7 + srv_n_purge_threads - 1: srv_worker_thread */
 static os_thread_id_t	thread_ids[SRV_MAX_N_IO_THREADS + 7
-				   + SRV_MAX_N_PURGE_THREADS];
+				   + SRV_MAX_N_PURGE_THREADS
+				   + MTFLUSH_MAX_WORKER];
+/* Thread contex data for multi-threaded flush */
+void *mtflush_ctx=NULL;
 
 /** Thead handles */
 static os_thread_t	thread_handles[SRV_MAX_N_IO_THREADS + 7 + SRV_MAX_N_PURGE_THREADS];
@@ -522,7 +531,8 @@ DECLARE_THREAD(io_handler_thread)(
 	os_thread_set_priority(srv_io_tids[tid_i], srv_sched_priority_io);
 
 #ifdef UNIV_DEBUG_THREAD_CREATION
-	fprintf(stderr, "Io handler thread %lu starts, id %lu\n", segment,
+	ib_logf(IB_LOG_LEVEL_INFO,
+		"Io handler thread %lu starts, id %lu\n", segment,
 		os_thread_pf(os_thread_get_curr_id()));
 #endif
 
@@ -573,7 +583,7 @@ static MY_ATTRIBUTE((nonnull, warn_unused_result))
 dberr_t
 create_log_file(
 /*============*/
-	pfs_os_file_t*	file,	/*!< out: file handle */
+	os_file_t*	file,	/*!< out: file handle */
 	const char*	name)	/*!< in: log file name */
 {
 	ibool		ret;
@@ -581,7 +591,7 @@ create_log_file(
 	*file = os_file_create(
 		innodb_file_log_key, name,
 		OS_FILE_CREATE|OS_FILE_ON_ERROR_NO_EXIT, OS_FILE_NORMAL,
-		OS_LOG_FILE, &ret);
+		OS_LOG_FILE, &ret, FALSE);
 
 	if (!ret) {
 		ib_logf(IB_LOG_LEVEL_ERROR, "Cannot create %s", name);
@@ -612,6 +622,19 @@ create_log_file(
 /** Initial number of the first redo log file */
 #define INIT_LOG_FILE0	(SRV_N_LOG_FILES_MAX + 1)
 
+#ifdef DBUG_OFF
+# define RECOVERY_CRASH(x) do {} while(0)
+#else
+# define RECOVERY_CRASH(x) do {						\
+	if (srv_force_recovery_crash == x) {				\
+		fprintf(stderr, "innodb_force_recovery_crash=%lu\n",	\
+			srv_force_recovery_crash);			\
+		fflush(stderr);						\
+		exit(3);						\
+	}								\
+} while (0)
+#endif
+
 /*********************************************************************//**
 Creates all log files.
 @return	DB_SUCCESS or error code */
@@ -652,14 +675,13 @@ create_log_files(
 			file should be recoverable. The buffer
 			pool was clean, and we can simply create
 			all log files from the scratch. */
-			DBUG_EXECUTE_IF("innodb_log_abort_6",
-					return(DB_ERROR););
+			RECOVERY_CRASH(6);
 		}
 	}
 
 	ut_ad(!buf_pool_check_no_pending_io());
 
-	DBUG_EXECUTE_IF("innodb_log_abort_7", return(DB_ERROR););
+	RECOVERY_CRASH(7);
 
 	for (unsigned i = 0; i < srv_n_log_files; i++) {
 		sprintf(logfilename + dirnamelen,
@@ -672,7 +694,7 @@ create_log_files(
 		}
 	}
 
-	DBUG_EXECUTE_IF("innodb_log_abort_8", return(DB_ERROR););
+	RECOVERY_CRASH(8);
 
 	/* We did not create the first log file initially as
 	ib_logfile0, so that crash recovery cannot find it until it
@@ -682,7 +704,8 @@ create_log_files(
 	fil_space_create(
 		logfilename, SRV_LOG_SPACE_FIRST_ID,
 		fsp_flags_set_page_size(0, UNIV_PAGE_SIZE),
-		FIL_LOG);
+		FIL_LOG,
+		NULL /* no encryption yet */);
 	ut_a(fil_validate());
 
 	logfile0 = fil_node_create(
@@ -704,7 +727,7 @@ create_log_files(
 #ifdef UNIV_LOG_ARCHIVE
 	/* Create the file space object for archived logs. */
 	fil_space_create("arch_log_space", SRV_LOG_SPACE_FIRST_ID + 1,
-			 0, FIL_LOG);
+		0, FIL_LOG, NULL /* no encryption yet */);
 #endif
 	log_group_init(0, srv_n_log_files,
 		       srv_log_file_size * UNIV_PAGE_SIZE,
@@ -727,16 +750,10 @@ create_log_files(
 	return(DB_SUCCESS);
 }
 
-/** Rename the first redo log file.
-@param[in,out]	logfilename	buffer for the log file name
-@param[in]	dirnamelen	length of the directory path
-@param[in]	lsn		FIL_PAGE_FILE_FLUSH_LSN value
-@param[in,out]	logfile0	name of the first log file
-@return	error code
-@retval	DB_SUCCESS	on successful operation */
-MY_ATTRIBUTE((warn_unused_result, nonnull))
+/*********************************************************************//**
+Renames the first log file. */
 static
-dberr_t
+void
 create_log_files_rename(
 /*====================*/
 	char*	logfilename,	/*!< in/out: buffer for log file name */
@@ -747,9 +764,6 @@ create_log_files_rename(
 	/* If innodb_flush_method=O_DSYNC,
 	we need to explicitly flush the log buffers. */
 	fil_flush(SRV_LOG_SPACE_FIRST_ID);
-
-	DBUG_EXECUTE_IF("innodb_log_abort_9", return(DB_ERROR););
-
 	/* Close the log files, so that we can rename
 	the first one. */
 	fil_close_log_files(false);
@@ -758,28 +772,26 @@ create_log_files_rename(
 	checkpoint has been created. */
 	sprintf(logfilename + dirnamelen, "ib_logfile%u", 0);
 
+	RECOVERY_CRASH(9);
+
 	ib_logf(IB_LOG_LEVEL_INFO,
 		"Renaming log file %s to %s", logfile0, logfilename);
 
 	mutex_enter(&log_sys->mutex);
 	ut_ad(strlen(logfile0) == 2 + strlen(logfilename));
-	dberr_t err = os_file_rename(
-		innodb_file_log_key, logfile0, logfilename)
-		? DB_SUCCESS : DB_ERROR;
+	ibool success = os_file_rename(
+		innodb_file_log_key, logfile0, logfilename);
+	ut_a(success);
+
+	RECOVERY_CRASH(10);
 
 	/* Replace the first file with ib_logfile0. */
 	strcpy(logfile0, logfilename);
 	mutex_exit(&log_sys->mutex);
 
-	DBUG_EXECUTE_IF("innodb_log_abort_10", err = DB_ERROR;);
+	fil_open_log_and_system_tablespace_files();
 
-	if (err == DB_SUCCESS) {
-		fil_open_log_and_system_tablespace_files();
-		ib_logf(IB_LOG_LEVEL_WARN,
-			"New log files created, LSN=" LSN_PF, lsn);
-	}
-
-	return(err);
+	ib_logf(IB_LOG_LEVEL_WARN, "New log files created, LSN=" LSN_PF, lsn);
 }
 
 /*********************************************************************//**
@@ -789,7 +801,7 @@ static MY_ATTRIBUTE((nonnull, warn_unused_result))
 dberr_t
 open_log_file(
 /*==========*/
-	pfs_os_file_t*	file,	/*!< out: file handle */
+	os_file_t*	file,	/*!< out: file handle */
 	const char*	name,	/*!< in: log file name */
 	os_offset_t*	size)	/*!< out: file size */
 {
@@ -797,7 +809,7 @@ open_log_file(
 
 	*file = os_file_create(innodb_file_log_key, name,
 			       OS_FILE_OPEN, OS_FILE_AIO,
-			       OS_LOG_FILE, &ret);
+			       OS_LOG_FILE, &ret, FALSE);
 	if (!ret) {
 		ib_logf(IB_LOG_LEVEL_ERROR, "Unable to open '%s'", name);
 		return(DB_ERROR);
@@ -841,6 +853,7 @@ open_or_create_data_files(
 	ulint		space;
 	ulint		rounded_size_pages;
 	char		name[10000];
+	fil_space_crypt_t*    crypt_data;
 
 	if (srv_n_data_files >= 1000) {
 
@@ -888,7 +901,7 @@ open_or_create_data_files(
 
 			files[i] = os_file_create(
 				innodb_file_data_key, name, OS_FILE_CREATE,
-				OS_FILE_NORMAL, OS_DATA_FILE, &ret);
+				OS_FILE_NORMAL, OS_DATA_FILE, &ret, FALSE);
 
 			if (srv_read_only_mode) {
 
@@ -905,7 +918,7 @@ open_or_create_data_files(
 				   && os_file_get_last_error(false)
 				   != OS_FILE_ALREADY_EXISTS
 #ifdef UNIV_AIX
-				   /* AIX 5.1 after security patch ML7 may have
+			    	   /* AIX 5.1 after security patch ML7 may have
 			           errno set to 0 here, which causes our
 				   function to return 100; work around that
 				   AIX problem */
@@ -931,7 +944,7 @@ open_or_create_data_files(
 
 			files[i] = os_file_create(
 				innodb_file_data_key, name, OS_FILE_OPEN_RAW,
-				OS_FILE_NORMAL, OS_DATA_FILE, &ret);
+				OS_FILE_NORMAL, OS_DATA_FILE, &ret, FALSE);
 
 			if (!ret) {
 				ib_logf(IB_LOG_LEVEL_ERROR,
@@ -943,7 +956,7 @@ open_or_create_data_files(
 			const char*	check_msg;
 			check_msg = fil_read_first_page(
 				files[i], FALSE, &flags, &space,
-				min_flushed_lsn, max_flushed_lsn);
+				min_flushed_lsn, max_flushed_lsn, NULL);
 
 			/* If first page is valid, don't overwrite DB.
 			It prevents overwriting DB when mysql_install_db
@@ -979,17 +992,17 @@ open_or_create_data_files(
 				files[i] = os_file_create(
 					innodb_file_data_key,
 					name, OS_FILE_OPEN_RAW,
-					OS_FILE_NORMAL, OS_DATA_FILE, &ret);
+					OS_FILE_NORMAL, OS_DATA_FILE, &ret, FALSE);
 			} else if (i == 0) {
 				files[i] = os_file_create(
 					innodb_file_data_key,
 					name, OS_FILE_OPEN_RETRY,
-					OS_FILE_NORMAL, OS_DATA_FILE, &ret);
+					OS_FILE_NORMAL, OS_DATA_FILE, &ret, FALSE);
 			} else {
 				files[i] = os_file_create(
 					innodb_file_data_key,
 					name, OS_FILE_OPEN, OS_FILE_NORMAL,
-					OS_DATA_FILE, &ret);
+					OS_DATA_FILE, &ret, FALSE);
 			}
 
 			if (!ret) {
@@ -1077,7 +1090,7 @@ skip_size_check:
 check_first_page:
 			check_msg = fil_read_first_page(
 				files[i], one_opened, &flags, &space,
-				min_flushed_lsn, max_flushed_lsn);
+				min_flushed_lsn, max_flushed_lsn, &crypt_data);
 
 			if (check_msg) {
 
@@ -1171,6 +1184,8 @@ check_first_page:
 			}
 
 			*sum_of_new_sizes += srv_data_file_sizes[i];
+
+			crypt_data = fil_space_create_crypt_data(FIL_SPACE_ENCRYPTION_DEFAULT, FIL_DEFAULT_ENCRYPTION_KEY);
 		}
 
 		ret = os_file_close(files[i]);
@@ -1178,7 +1193,9 @@ check_first_page:
 
 		if (i == 0) {
 			flags = fsp_flags_set_page_size(0, UNIV_PAGE_SIZE);
-			fil_space_create(name, 0, flags, FIL_TABLESPACE);
+			fil_space_create(name, 0, flags, FIL_TABLESPACE,
+					 crypt_data);
+			crypt_data = NULL;
 		}
 
 		ut_a(fil_validate());
@@ -1202,7 +1219,7 @@ srv_undo_tablespace_create(
 	const char*	name,		/*!< in: tablespace name */
 	ulint		size)		/*!< in: tablespace size in pages */
 {
-	pfs_os_file_t	fh;
+	os_file_t	fh;
 	ibool		ret;
 	dberr_t		err = DB_SUCCESS;
 
@@ -1212,7 +1229,7 @@ srv_undo_tablespace_create(
 		innodb_file_data_key,
 		name,
 		srv_read_only_mode ? OS_FILE_OPEN : OS_FILE_CREATE,
-		OS_FILE_NORMAL, OS_DATA_FILE, &ret);
+		OS_FILE_NORMAL, OS_DATA_FILE, &ret, FALSE);
 
 	if (srv_read_only_mode && ret) {
 		ib_logf(IB_LOG_LEVEL_INFO,
@@ -1279,7 +1296,7 @@ srv_undo_tablespace_open(
 	const char*	name,		/*!< in: tablespace name */
 	ulint		space)		/*!< in: tablespace id */
 {
-	pfs_os_file_t	fh;
+	os_file_t	fh;
 	dberr_t		err	= DB_ERROR;
 	ibool		ret;
 	ulint		flags;
@@ -1299,7 +1316,8 @@ srv_undo_tablespace_open(
 		| OS_FILE_ON_ERROR_SILENT,
 		OS_FILE_NORMAL,
 		OS_DATA_FILE,
-		&ret);
+		&ret,
+		FALSE);
 
 	/* If the file open was successful then load the tablespace. */
 
@@ -1323,7 +1341,8 @@ srv_undo_tablespace_open(
 
 		/* Set the compressed page size to 0 (non-compressed) */
 		flags = fsp_flags_set_page_size(0, UNIV_PAGE_SIZE);
-		fil_space_create(name, space, flags, FIL_TABLESPACE);
+		fil_space_create(name, space, flags, FIL_TABLESPACE,
+				 NULL /* no encryption */);
 
 		ut_a(fil_validate());
 
@@ -1378,23 +1397,13 @@ srv_undo_tablespaces_init(
 
 	for (i = 0; create_new_db && i < n_conf_tablespaces; ++i) {
 		char	name[OS_FILE_MAX_PATH];
-		ulint	space_id  = i + 1;
-
-		DBUG_EXECUTE_IF("innodb_undo_upgrade",
-				space_id = i + 3;);
 
 		ut_snprintf(
 			name, sizeof(name),
 			"%s%cundo%03lu",
-			srv_undo_dir, SRV_PATH_SEPARATOR, space_id);
-
-		if (i == 0) {
-			srv_undo_space_id_start = space_id;
-			prev_space_id = srv_undo_space_id_start - 1;
-		}
-
-		undo_tablespace_ids[i] = space_id;
+			srv_undo_dir, SRV_PATH_SEPARATOR, i + 1);
 
+		/* Undo space ids start from 1. */
 		err = srv_undo_tablespace_create(
 			name, SRV_UNDO_TABLESPACE_SIZE_IN_PAGES);
 
@@ -1416,16 +1425,14 @@ srv_undo_tablespaces_init(
 	if (!create_new_db) {
 		n_undo_tablespaces = trx_rseg_get_n_undo_tablespaces(
 			undo_tablespace_ids);
-
-		if (n_undo_tablespaces != 0) {
-			srv_undo_space_id_start = undo_tablespace_ids[0];
-			prev_space_id = srv_undo_space_id_start - 1;
-		}
-
 	} else {
 		n_undo_tablespaces = n_conf_tablespaces;
 
-		undo_tablespace_ids[n_conf_tablespaces] = ULINT_UNDEFINED;
+		for (i = 1; i <= n_undo_tablespaces; ++i) {
+			undo_tablespace_ids[i - 1] = i;
+		}
+
+		undo_tablespace_ids[i] = ULINT_UNDEFINED;
 	}
 
 	/* Open all the undo tablespaces that are currently in use. If we
@@ -1449,6 +1456,8 @@ srv_undo_tablespaces_init(
 		ut_a(undo_tablespace_ids[i] != 0);
 		ut_a(undo_tablespace_ids[i] != ULINT_UNDEFINED);
 
+		/* Undo space ids start from 1. */
+
 		err = srv_undo_tablespace_open(name, undo_tablespace_ids[i]);
 
 		if (err != DB_SUCCESS) {
@@ -1483,23 +1492,11 @@ srv_undo_tablespaces_init(
 			break;
 		}
 
-		/** Note the first undo tablespace id in case of
-		no active undo tablespace. */
-		if (n_undo_tablespaces == 0) {
-			srv_undo_space_id_start = i;
-		}
-
 		++n_undo_tablespaces;
 
 		++*n_opened;
 	}
 
-	/** Explictly specify the srv_undo_space_id_start
-	as zero when there are no undo tablespaces. */
-	if (n_undo_tablespaces == 0) {
-		srv_undo_space_id_start = 0;
-	}
-
 	/* If the user says that there are fewer than what we find we
 	tolerate that discrepancy but not the inverse. Because there could
 	be unused undo tablespaces for future use. */
@@ -1544,11 +1541,10 @@ srv_undo_tablespaces_init(
 		mtr_start(&mtr);
 
 		/* The undo log tablespace */
-		for (i = 0; i < n_undo_tablespaces; ++i) {
+		for (i = 1; i <= n_undo_tablespaces; ++i) {
 
 			fsp_header_init(
-				undo_tablespace_ids[i],
-				SRV_UNDO_TABLESPACE_SIZE_IN_PAGES, &mtr);
+				i, SRV_UNDO_TABLESPACE_SIZE_IN_PAGES, &mtr);
 		}
 
 		mtr_commit(&mtr);
@@ -1634,6 +1630,8 @@ innobase_start_or_create_for_mysql(void)
 	lsn_t		max_arch_log_no	= LSN_MAX;
 #endif /* UNIV_LOG_ARCHIVE */
 	ulint		sum_of_new_sizes;
+	ulint		sum_of_data_file_sizes;
+	ulint		tablespace_size_in_header;
 	dberr_t		err;
 	unsigned	i;
 	ulint		srv_n_log_files_found = srv_n_log_files;
@@ -1646,22 +1644,8 @@ innobase_start_or_create_for_mysql(void)
 	size_t		dirnamelen;
 	bool		sys_datafiles_created = false;
 
-	/* Check that os_fast_mutexes work as expected */
-	os_fast_mutex_init(PFS_NOT_INSTRUMENTED, &srv_os_test_mutex);
-
-	ut_a(0 == os_fast_mutex_trylock(&srv_os_test_mutex));
-
-	os_fast_mutex_unlock(&srv_os_test_mutex);
-
-	os_fast_mutex_lock(&srv_os_test_mutex);
-
-	os_fast_mutex_unlock(&srv_os_test_mutex);
-
-	os_fast_mutex_free(&srv_os_test_mutex);
-
-	if (srv_force_recovery == SRV_FORCE_NO_LOG_REDO) {
-		srv_read_only_mode = 1;
-	}
+	/* This should be initialized early */
+	ut_init_timer();
 
 	high_level_read_only = srv_read_only_mode
 		|| srv_force_recovery > SRV_FORCE_NO_TRX_UNDO;
@@ -1705,7 +1689,22 @@ innobase_start_or_create_for_mysql(void)
 #endif /* PAGE_ATOMIC_REF_COUNT */
 	);
 
-	compile_time_assert(sizeof(ulint) == sizeof(void*));
+
+	if (sizeof(ulint) != sizeof(void*)) {
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+			" InnoDB: Error: size of InnoDB's ulint is %lu, "
+			"but size of void*\n", (ulong) sizeof(ulint));
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+			" InnoDB: is %lu. The sizes should be the same "
+			"so that on a 64-bit\n",
+			(ulong) sizeof(void*));
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+			" InnoDB: platforms you can allocate more than 4 GB "
+			"of memory.\n");
+	}
 
 	/* If stacktrace is used we set up signal handler for SIGUSR2 signal
 	here. If signal handler set fails we report that and disable
@@ -1729,53 +1728,45 @@ innobase_start_or_create_for_mysql(void)
 	}
 
 #ifdef UNIV_DEBUG
-	ut_print_timestamp(stderr);
-	fprintf(stderr,
-		" InnoDB: !!!!!!!! UNIV_DEBUG switched on !!!!!!!!!\n");
+	ib_logf(IB_LOG_LEVEL_INFO,
+		" InnoDB: !!!!!!!! UNIV_DEBUG switched on !!!!!!!!!");
 #endif
 
 #ifdef UNIV_IBUF_DEBUG
-	ut_print_timestamp(stderr);
-	fprintf(stderr,
-		" InnoDB: !!!!!!!! UNIV_IBUF_DEBUG switched on !!!!!!!!!\n");
+	ib_logf(IB_LOG_LEVEL_INFO,
+		" InnoDB: !!!!!!!! UNIV_IBUF_DEBUG switched on !!!!!!!!!");
 # ifdef UNIV_IBUF_COUNT_DEBUG
-	ut_print_timestamp(stderr);
-	fprintf(stderr,
+	ib_logf(IB_LOG_LEVEL_INFO,
 		" InnoDB: !!!!!!!! UNIV_IBUF_COUNT_DEBUG switched on "
-		"!!!!!!!!!\n");
-	ut_print_timestamp(stderr);
-	fprintf(stderr,
-		" InnoDB: Crash recovery will fail with UNIV_IBUF_COUNT_DEBUG\n");
+		"!!!!!!!!!");
+	ib_logf(IB_LOG_LEVEL_INFO,
+		" InnoDB: Crash recovery will fail with UNIV_IBUF_COUNT_DEBUG");
 # endif
 #endif
 
 #ifdef UNIV_BLOB_DEBUG
-	fprintf(stderr,
+	ib_logf(IB_LOG_LEVEL_INFO,
 		"InnoDB: !!!!!!!! UNIV_BLOB_DEBUG switched on !!!!!!!!!\n"
-		"InnoDB: Server restart may fail with UNIV_BLOB_DEBUG\n");
+		"InnoDB: Server restart may fail with UNIV_BLOB_DEBUG");
 #endif /* UNIV_BLOB_DEBUG */
 
 #ifdef UNIV_SYNC_DEBUG
-	ut_print_timestamp(stderr);
-	fprintf(stderr,
-		" InnoDB: !!!!!!!! UNIV_SYNC_DEBUG switched on !!!!!!!!!\n");
+	ib_logf(IB_LOG_LEVEL_INFO,
+		" InnoDB: !!!!!!!! UNIV_SYNC_DEBUG switched on !!!!!!!!!");
 #endif
 
 #ifdef UNIV_SEARCH_DEBUG
-	ut_print_timestamp(stderr);
-	fprintf(stderr,
-		" InnoDB: !!!!!!!! UNIV_SEARCH_DEBUG switched on !!!!!!!!!\n");
+	ib_logf(IB_LOG_LEVEL_INFO,
+		" InnoDB: !!!!!!!! UNIV_SEARCH_DEBUG switched on !!!!!!!!!");
 #endif
 
 #ifdef UNIV_LOG_LSN_DEBUG
-	ut_print_timestamp(stderr);
-	fprintf(stderr,
-		" InnoDB: !!!!!!!! UNIV_LOG_LSN_DEBUG switched on !!!!!!!!!\n");
+	ib_logf(IB_LOG_LEVEL_INFO,
+		" InnoDB: !!!!!!!! UNIV_LOG_LSN_DEBUG switched on !!!!!!!!!");
 #endif /* UNIV_LOG_LSN_DEBUG */
 #ifdef UNIV_MEM_DEBUG
-	ut_print_timestamp(stderr);
-	fprintf(stderr,
-		" InnoDB: !!!!!!!! UNIV_MEM_DEBUG switched on !!!!!!!!!\n");
+	ib_logf(IB_LOG_LEVEL_INFO,
+		" InnoDB: !!!!!!!! UNIV_MEM_DEBUG switched on !!!!!!!!!");
 #endif
 
 	if (srv_use_sys_malloc) {
@@ -1845,32 +1836,7 @@ innobase_start_or_create_for_mysql(void)
 	srv_startup_is_before_trx_rollback_phase = TRUE;
 
 #ifdef __WIN__
-	switch (os_get_os_version()) {
-	case OS_WIN95:
-	case OS_WIN31:
-	case OS_WINNT:
-		srv_use_native_conditions = FALSE;
-		/* On Win 95, 98, ME, Win32 subsystem for Windows 3.1,
-		and NT use simulated aio. In NT Windows provides async i/o,
-		but when run in conjunction with InnoDB Hot Backup, it seemed
-		to corrupt the data files. */
-
-		srv_use_native_aio = FALSE;
-		break;
-
-	case OS_WIN2000:
-	case OS_WINXP:
-		/* On 2000 and XP, async IO is available, but no condition variables. */
-		srv_use_native_aio = TRUE;
-		srv_use_native_conditions = FALSE;
- 		break;
-
-	default:
-		/* Vista and later have both async IO and condition variables */
-		srv_use_native_aio = TRUE;
-		srv_use_native_conditions = TRUE;
-		break;
-	}
+  srv_use_native_aio = TRUE;
 
 #elif defined(LINUX_NATIVE_AIO)
 
@@ -1972,9 +1938,13 @@ innobase_start_or_create_for_mysql(void)
 
 	srv_boot();
 
-	ib_logf(IB_LOG_LEVEL_INFO,
-		"%s CPU crc32 instructions",
-		ut_crc32_sse2_enabled ? "Using" : "Not using");
+	if (ut_crc32_sse2_enabled) {
+		ib_logf(IB_LOG_LEVEL_INFO, "Using SSE crc32 instructions");
+	} else if (ut_crc32_power8_enabled) {
+		ib_logf(IB_LOG_LEVEL_INFO, "Using POWER8 crc32 instructions");
+	} else {
+		ib_logf(IB_LOG_LEVEL_INFO, "Using generic crc32 instructions");
+	}
 
 	if (!srv_read_only_mode) {
 
@@ -2129,7 +2099,6 @@ innobase_start_or_create_for_mysql(void)
 
 	fsp_init();
 	log_init();
-	log_online_init();
 
 	lock_sys_create(srv_lock_table_size);
 
@@ -2304,18 +2273,14 @@ innobase_start_or_create_for_mysql(void)
 						dirnamelen, max_flushed_lsn,
 						logfile0);
 
-					if (err == DB_SUCCESS) {
-						err = create_log_files_rename(
-							logfilename,
-							dirnamelen,
-							max_flushed_lsn,
-							logfile0);
-					}
-
 					if (err != DB_SUCCESS) {
 						return(err);
 					}
 
+					create_log_files_rename(
+						logfilename, dirnamelen,
+						max_flushed_lsn, logfile0);
+
 					/* Suppress the message about
 					crash recovery. */
 					max_flushed_lsn = min_flushed_lsn
@@ -2380,7 +2345,8 @@ innobase_start_or_create_for_mysql(void)
 		fil_space_create(logfilename,
 				 SRV_LOG_SPACE_FIRST_ID,
 				 fsp_flags_set_page_size(0, UNIV_PAGE_SIZE),
-				 FIL_LOG);
+				 FIL_LOG,
+				 NULL /* no encryption yet */);
 
 		ut_a(fil_validate());
 
@@ -2402,7 +2368,7 @@ innobase_start_or_create_for_mysql(void)
 		/* Create the file space object for archived logs. Under
 		MySQL, no archiving ever done. */
 		fil_space_create("arch_log_space", SRV_LOG_SPACE_FIRST_ID + 1,
-				 0, FIL_LOG);
+			0, FIL_LOG, NULL /* no encryption yet */);
 #endif /* UNIV_LOG_ARCHIVE */
 		log_group_init(0, i, srv_log_file_size * UNIV_PAGE_SIZE,
 			       SRV_LOG_SPACE_FIRST_ID,
@@ -2436,11 +2402,17 @@ files_checked:
 		dict_stats_thread_init();
 	}
 
+	if (!srv_read_only_mode && srv_scrub_log) {
+		/* TODO(minliz): have/use log_scrub_thread_init() instead? */
+		log_scrub_event = os_event_create();
+	}
+
 	trx_sys_file_format_init();
 
 	trx_sys_create();
 
 	if (create_new_db) {
+
 		ut_a(!srv_read_only_mode);
 		init_log_online();
 
@@ -2484,12 +2456,8 @@ files_checked:
 
 		fil_flush_file_spaces(FIL_TABLESPACE);
 
-		err = create_log_files_rename(logfilename, dirnamelen,
-					      max_flushed_lsn, logfile0);
-
-		if (err != DB_SUCCESS) {
-			return(err);
-		}
+		create_log_files_rename(logfilename, dirnamelen,
+					max_flushed_lsn, logfile0);
 	} else {
 
 		/* Check if we support the max format that is stamped
@@ -2518,22 +2486,6 @@ files_checked:
 		and there must be no page in the buf_flush list. */
 		buf_pool_invalidate();
 
-		/* Start monitor thread early enough so that e.g. crash
-		recovery failing to find free pages in the buffer pool is
-		diagnosed. */
-		if (!srv_read_only_mode)
-		{
-			/* Create the thread which prints InnoDB monitor
-			info */
-			thread_handles[4 + SRV_MAX_N_IO_THREADS] =
-				os_thread_create(
-					srv_monitor_thread,
-					NULL,
-					thread_ids + 4 + SRV_MAX_N_IO_THREADS);
-
-			thread_started[4 + SRV_MAX_N_IO_THREADS] = true;
-		}
-
 		/* We always try to do a recovery, even if the database had
 		been shut down normally: this is the normal startup path */
 
@@ -2542,92 +2494,25 @@ files_checked:
 			min_flushed_lsn, max_flushed_lsn);
 
 		if (err != DB_SUCCESS) {
-			return(err);
+
+			return(DB_ERROR);
 		}
 
 		init_log_online();
 
-		/* Initialize the change buffer. */
+		/* Since the insert buffer init is in dict_boot, and the
+		insert buffer is needed in any disk i/o, first we call
+		dict_boot(). Note that trx_sys_init_at_db_start() only needs
+		to access space 0, and the insert buffer at this stage already
+		works for space 0. */
+
 		err = dict_boot();
 
 		if (err != DB_SUCCESS) {
 			return(err);
 		}
 
-		if (!srv_read_only_mode) {
-			if (sum_of_new_sizes > 0) {
-				/* New data file(s) were added */
-				mtr_start(&mtr);
-				fsp_header_inc_size(0, sum_of_new_sizes, &mtr);
-				mtr_commit(&mtr);
-				/* Immediately write the log record about
-				increased tablespace size to disk, so that it
-				is durable even if mysqld would crash
-				quickly */
-				log_buffer_flush_to_disk();
-			}
-		}
-
-		const ulint	tablespace_size_in_header
-			= fsp_header_get_tablespace_size();
-
-#ifdef UNIV_DEBUG
-		/* buf_debug_prints = TRUE; */
-#endif /* UNIV_DEBUG */
-		ulint	sum_of_data_file_sizes = 0;
-
-		for (ulint d = 0; d < srv_n_data_files; d++) {
-			sum_of_data_file_sizes += srv_data_file_sizes[d];
-		}
-
-		/* Compare the system tablespace file size to what is
-		stored in FSP_SIZE. In open_or_create_data_files()
-		we already checked that the file sizes match the
-		innodb_data_file_path specification. */
-		if (srv_read_only_mode
-		    || sum_of_data_file_sizes == tablespace_size_in_header) {
-			/* Do not complain about the size. */
-		} else if (!srv_auto_extend_last_data_file
-			   || sum_of_data_file_sizes
-			   < tablespace_size_in_header) {
-			ib_logf(IB_LOG_LEVEL_ERROR,
-				"Tablespace size stored in header is " ULINTPF
-				" pages, but the sum of data file sizes is "
-				ULINTPF " pages",
-				tablespace_size_in_header,
-				sum_of_data_file_sizes);
-
-			if (srv_force_recovery == 0
-			    && sum_of_data_file_sizes
-			    < tablespace_size_in_header) {
-				ib_logf(IB_LOG_LEVEL_ERROR,
-					"Cannot start InnoDB. The tail of"
-					" the system tablespace is"
-					" missing. Have you edited"
-					" innodb_data_file_path in my.cnf"
-					" in an inappropriate way, removing"
-					" data files from there?"
-					" You can set innodb_force_recovery=1"
-					" in my.cnf to force"
-					" a startup if you are trying to"
-					" recover a badly corrupt database.");
-
-				return(DB_ERROR);
-			}
-		}
-
-		/* This must precede recv_apply_hashed_log_recs(true). */
 		ib_bh = trx_sys_init_at_db_start();
-
-		if (srv_force_recovery < SRV_FORCE_NO_LOG_REDO) {
-			/* Apply the hashed log records to the
-			respective file pages, for the last batch of
-			recv_group_scan_log_recs(). */
-
-			recv_apply_hashed_log_recs(true);
-			DBUG_PRINT("ib_log", ("apply completed"));
-		}
-
 		n_recovered_trx = UT_LIST_GET_LEN(trx_sys->rw_trx_list);
 
 		/* The purge system needs to create the purge view and
@@ -2696,8 +2581,7 @@ files_checked:
 				ULINT_MAX, LSN_MAX, NULL);
 			ut_a(success);
 
-			DBUG_EXECUTE_IF("innodb_log_abort_1",
-					return(DB_ERROR););
+			RECOVERY_CRASH(1);
 
 			min_flushed_lsn = max_flushed_lsn = log_get_lsn();
 
@@ -2712,6 +2596,8 @@ files_checked:
 
 			buf_flush_wait_batch_end(NULL, BUF_FLUSH_LIST);
 
+			RECOVERY_CRASH(2);
+
 			/* Flush the old log files. */
 			log_buffer_flush_to_disk();
 			/* If innodb_flush_method=O_DSYNC,
@@ -2726,27 +2612,21 @@ files_checked:
 			ut_d(recv_no_log_write = TRUE);
 			ut_ad(!buf_pool_check_no_pending_io());
 
-			DBUG_EXECUTE_IF("innodb_log_abort_3",
-					return(DB_ERROR););
+			RECOVERY_CRASH(3);
 
 			/* Stamp the LSN to the data files. */
 			fil_write_flushed_lsn_to_data_files(
 				max_flushed_lsn, 0);
 
-			DBUG_EXECUTE_IF("innodb_log_abort_4", err = DB_ERROR;);
-
-			if (err != DB_SUCCESS) {
-				return(err);
-			}
-
 			fil_flush_file_spaces(FIL_TABLESPACE);
 
+			RECOVERY_CRASH(4);
+
 			/* Close and free the redo log files, so that
 			we can replace them. */
 			fil_close_log_files(true);
 
-			DBUG_EXECUTE_IF("innodb_log_abort_5",
-					return(DB_ERROR););
+			RECOVERY_CRASH(5);
 
 			/* Free the old log file space. */
 			log_group_close_all();
@@ -2770,11 +2650,8 @@ files_checked:
 			fil_write_flushed_lsn_to_data_files(min_flushed_lsn, 0);
 			fil_flush_file_spaces(FIL_TABLESPACE);
 
-			err = create_log_files_rename(logfilename, dirnamelen,
-						      log_get_lsn(), logfile0);
-			if (err != DB_SUCCESS) {
-				return(err);
-			}
+			create_log_files_rename(logfilename, dirnamelen,
+						log_get_lsn(), logfile0);
 		}
 
 		srv_startup_is_before_trx_rollback_phase = FALSE;
@@ -2788,8 +2665,20 @@ files_checked:
 		trx_sys_file_format_tag_init();
 	}
 
-	ut_ad(err == DB_SUCCESS);
-	ut_a(sum_of_new_sizes != ULINT_UNDEFINED);
+	if (!create_new_db && sum_of_new_sizes > 0) {
+		/* New data file(s) were added */
+		mtr_start(&mtr);
+
+		fsp_header_inc_size(0, sum_of_new_sizes, &mtr);
+
+		mtr_commit(&mtr);
+
+		/* Immediately write the log record about increased tablespace
+		size to disk, so that it is durable even if mysqld would crash
+		quickly */
+
+		log_buffer_flush_to_disk();
+	}
 
 #ifdef UNIV_LOG_ARCHIVE
 	if (!srv_read_only_mode) {
@@ -2868,13 +2757,10 @@ files_checked:
 		thread_started[3 + SRV_MAX_N_IO_THREADS] = true;
 
 		/* Create the thread which prints InnoDB monitor info */
-		if (!thread_started[4 + SRV_MAX_N_IO_THREADS]) {
-			/* srv_monitor_thread not yet started */
-			thread_handles[4 + SRV_MAX_N_IO_THREADS] = os_thread_create(
-				srv_monitor_thread,
-				NULL, thread_ids + 4 + SRV_MAX_N_IO_THREADS);
-			thread_started[4 + SRV_MAX_N_IO_THREADS] = true;
-		}
+		thread_handles[4 + SRV_MAX_N_IO_THREADS] = os_thread_create(
+			srv_monitor_thread,
+			NULL, thread_ids + 4 + SRV_MAX_N_IO_THREADS);
+		thread_started[4 + SRV_MAX_N_IO_THREADS] = true;
 	}
 
 	/* Create the SYS_FOREIGN and SYS_FOREIGN_COLS system tables */
@@ -2934,6 +2820,21 @@ files_checked:
 	}
 
 	if (!srv_read_only_mode) {
+
+		if (srv_use_mtflush) {
+			/* Start multi-threaded flush threads */
+			mtflush_ctx = buf_mtflu_handler_init(
+				srv_mtflush_threads,
+				srv_buf_pool_instances);
+
+			/* Set up the thread ids */
+			buf_mtflu_set_thread_ids(
+				srv_mtflush_threads,
+				mtflush_ctx,
+				(thread_ids + 6 + SRV_MAX_N_PURGE_THREADS));
+		}
+
+
 		buf_flush_page_cleaner_thread_handle = os_thread_create(buf_flush_page_cleaner_thread, NULL, NULL);
 		buf_flush_page_cleaner_thread_started = true;
 	}
@@ -2941,6 +2842,125 @@ files_checked:
 	buf_flush_lru_manager_thread_handle = os_thread_create(buf_flush_lru_manager_thread, NULL, NULL);
 	buf_flush_lru_manager_thread_started = true;
 
+#ifdef UNIV_DEBUG
+	/* buf_debug_prints = TRUE; */
+#endif /* UNIV_DEBUG */
+	sum_of_data_file_sizes = 0;
+
+	for (i = 0; i < srv_n_data_files; i++) {
+		sum_of_data_file_sizes += srv_data_file_sizes[i];
+	}
+
+	tablespace_size_in_header = fsp_header_get_tablespace_size();
+
+	if (!srv_read_only_mode
+	    && !srv_auto_extend_last_data_file
+	    && sum_of_data_file_sizes != tablespace_size_in_header) {
+
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+			" InnoDB: Error: tablespace size"
+			" stored in header is %lu pages, but\n",
+			(ulong) tablespace_size_in_header);
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+			"InnoDB: the sum of data file sizes is %lu pages\n",
+			(ulong) sum_of_data_file_sizes);
+
+		if (srv_force_recovery == 0
+		    && sum_of_data_file_sizes < tablespace_size_in_header) {
+			/* This is a fatal error, the tail of a tablespace is
+			missing */
+
+			ut_print_timestamp(stderr);
+			fprintf(stderr,
+				" InnoDB: Cannot start InnoDB."
+				" The tail of the system tablespace is\n");
+			ut_print_timestamp(stderr);
+			fprintf(stderr,
+				" InnoDB: missing. Have you edited"
+				" innodb_data_file_path in my.cnf in an\n");
+			ut_print_timestamp(stderr);
+			fprintf(stderr,
+				" InnoDB: inappropriate way, removing"
+				" ibdata files from there?\n");
+			ut_print_timestamp(stderr);
+			fprintf(stderr,
+				" InnoDB: You can set innodb_force_recovery=1"
+				" in my.cnf to force\n");
+			ut_print_timestamp(stderr);
+			fprintf(stderr,
+				" InnoDB: a startup if you are trying"
+				" to recover a badly corrupt database.\n");
+
+			return(DB_ERROR);
+		}
+	}
+
+	if (!srv_read_only_mode
+	    && srv_auto_extend_last_data_file
+	    && sum_of_data_file_sizes < tablespace_size_in_header) {
+
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+			" InnoDB: Error: tablespace size stored in header"
+			" is %lu pages, but\n",
+			(ulong) tablespace_size_in_header);
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+			" InnoDB: the sum of data file sizes"
+			" is only %lu pages\n",
+			(ulong) sum_of_data_file_sizes);
+
+		if (srv_force_recovery == 0) {
+
+			ut_print_timestamp(stderr);
+			fprintf(stderr,
+				" InnoDB: Cannot start InnoDB. The tail of"
+				" the system tablespace is\n");
+			ut_print_timestamp(stderr);
+			fprintf(stderr,
+				" InnoDB: missing. Have you edited"
+				" innodb_data_file_path in my.cnf in an\n");
+			ut_print_timestamp(stderr);
+			fprintf(stderr,
+				" InnoDB: inappropriate way, removing"
+				" ibdata files from there?\n");
+			ut_print_timestamp(stderr);
+			fprintf(stderr,
+				" InnoDB: You can set innodb_force_recovery=1"
+				" in my.cnf to force\n");
+			ut_print_timestamp(stderr);
+			fprintf(stderr,
+				" InnoDB: a startup if you are trying to"
+				" recover a badly corrupt database.\n");
+
+			return(DB_ERROR);
+		}
+	}
+
+	/* Check that os_fast_mutexes work as expected */
+	os_fast_mutex_init(PFS_NOT_INSTRUMENTED, &srv_os_test_mutex);
+
+	if (0 != os_fast_mutex_trylock(&srv_os_test_mutex)) {
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+			" InnoDB: Error: pthread_mutex_trylock returns"
+			" an unexpected value on\n");
+		ut_print_timestamp(stderr);
+		fprintf(stderr,
+			" InnoDB: success! Cannot continue.\n");
+		exit(1);
+	}
+
+	os_fast_mutex_unlock(&srv_os_test_mutex);
+
+	os_fast_mutex_lock(&srv_os_test_mutex);
+
+	os_fast_mutex_unlock(&srv_os_test_mutex);
+
+	os_fast_mutex_free(&srv_os_test_mutex);
+
 	if (!srv_file_per_table && srv_pass_corrupt_table) {
 		fprintf(stderr, "InnoDB: Warning:"
 			" The option innodb_file_per_table is disabled,"
@@ -2960,6 +2980,15 @@ files_checked:
 			(ulong) srv_force_recovery);
 	}
 
+	if (!srv_read_only_mode) {
+		/*
+		  Create a checkpoint before logging anything new, so that
+		  the current encryption key in use is definitely logged
+		  before any log blocks encrypted with that key.
+		*/
+		log_make_checkpoint_at(LSN_MAX, TRUE);
+	}
+
 	if (srv_force_recovery == 0) {
 		/* In the insert buffer we may have even bigger tablespace
 		id's, because we may have dropped those tablespaces, but
@@ -2970,9 +2999,24 @@ files_checked:
 	}
 
 	if (!srv_read_only_mode) {
+#ifdef WITH_WSREP
+		/*
+		  Create the dump/load thread only when not running with
+		  --wsrep-recover.
+		*/
+		if (!wsrep_recovery) {
+#endif /* WITH_WSREP */
 		/* Create the buffer pool dump/load thread */
-		buf_dump_thread_handle = os_thread_create(buf_dump_thread, NULL, NULL);
+		buf_dump_thread_handle=
+			os_thread_create(buf_dump_thread, NULL, NULL);
 		buf_dump_thread_started = true;
+#ifdef WITH_WSREP
+		} else {
+			ib_logf(IB_LOG_LEVEL_WARN,
+				"Skipping buffer pool dump/restore during "
+				"wsrep recovery.");
+		}
+#endif /* WITH_WSREP */
 
 		/* Create the dict stats gathering thread */
 		dict_stats_thread_handle = os_thread_create(dict_stats_thread, NULL, NULL);
@@ -2980,8 +3024,23 @@ files_checked:
 
 		/* Create the thread that will optimize the FTS sub-system. */
 		fts_optimize_init();
+
+		/* Create thread(s) that handles key rotation */
+		fil_system_enter();
+		fil_crypt_threads_init();
+		fil_system_exit();
+
+		/* Create the log scrub thread */
+		if (srv_scrub_log)
+			os_thread_create(log_scrub_thread, NULL, NULL);
 	}
 
+	/* Init data for datafile scrub threads */
+	btr_scrub_init();
+
+	/* Initialize online defragmentation. */
+	btr_defragment_init();
+
 	srv_was_started = TRUE;
 
 	return(DB_SUCCESS);
@@ -3042,6 +3101,9 @@ innobase_shutdown_for_mysql(void)
 		fts_optimize_start_shutdown();
 
 		fts_optimize_end();
+
+		/* Shutdown key rotation threads */
+		fil_crypt_threads_end();
 	}
 
 	/* 1. Flush the buffer pool to disk, write the current lsn to
@@ -3092,6 +3154,12 @@ innobase_shutdown_for_mysql(void)
 		logs_empty_and_mark_files_at_shutdown() and should have
 		already quit or is quitting right now. */
 
+		if (srv_use_mtflush) {
+			/* g. Exit the multi threaded flush threads */
+
+			buf_mtflu_io_thread_exit();
+		}
+
 		os_rmb;
 		if (os_thread_count == 0) {
 			/* All the threads have exited or are just exiting;
@@ -3138,8 +3206,20 @@ innobase_shutdown_for_mysql(void)
 
 	if (!srv_read_only_mode) {
 		dict_stats_thread_deinit();
+		if (srv_scrub_log) {
+			/* TODO(minliz): have/use log_scrub_thread_deinit() instead? */
+			os_event_free(log_scrub_event);
+			log_scrub_event = NULL;
+		}
 	}
 
+	if (!srv_read_only_mode) {
+		fil_crypt_threads_cleanup();
+	}
+
+	/* Cleanup data for datafile scrubbing */
+	btr_scrub_cleanup();
+
 #ifdef __WIN__
 	/* MDEV-361: ha_innodb.dll leaks handles on Windows
 	MDEV-7403: should not pass recv_writer_thread_handle to
@@ -3181,7 +3261,6 @@ innobase_shutdown_for_mysql(void)
 	btr_search_disable();
 
 	ibuf_close();
-	log_online_shutdown();
 	log_shutdown();
 	trx_sys_file_format_close();
 	trx_sys_close();
